# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import mock
import pytest


def test_should_register_a_state_machine(caplog):
    from statemachine import StateMachine, State, registry

    class CampaignMachine(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True)
        producing = State("Being produced")

        add_job = draft.to(draft) | producing.to(producing)
        produce = draft.to(producing)

    assert "CampaignMachine" in registry._REGISTRY
    assert (
        registry.get_machine_cls("tests.test_registry.CampaignMachine")
        == CampaignMachine
    )

    with pytest.warns(DeprecationWarning):
        assert registry.get_machine_cls("CampaignMachine") == CampaignMachine


@pytest.fixture()
def django_autodiscover_modules():
    import sys

    real_django = sys.modules.get("django")

    django = mock.MagicMock()
    module_loading = mock.MagicMock()
    auto_discover_modules = module_loading.autodiscover_modules

    sys.modules["django"] = django
    sys.modules["django.utils.module_loading"] = module_loading

    with mock.patch(
        "statemachine.registry._autodiscover_modules", new=auto_discover_modules
    ):
        yield auto_discover_modules

    del sys.modules["django"]
    del sys.modules["django.utils.module_loading"]
    if real_django:
        sys.modules["django"] = real_django


def test_load_modules_should_call_autodiscover_modules(django_autodiscover_modules):
    from statemachine.registry import load_modules

    # given
    modules = ["a", "c", "statemachine", "statemachines"]

    # when
    load_modules(modules)

    # then
    django_autodiscover_modules.assert_has_calls(mock.call(m) for m in modules)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import warnings

try:
    from django.utils.translation import ugettext
except Exception:

    def ugettext(text):
        return text


def qualname(cls):
    """
    Returns a fully qualified name of the class, to avoid name collisions.
    """
    return ".".join([cls.__module__, cls.__name__])


def _is_string(obj):
    return isinstance(obj, (str, type("")))  # type(u""") is a small hack for Python2


def ensure_iterable(obj):
    if _is_string(obj):
        return [obj]
    try:
        return iter(obj)
    except TypeError:
        return [obj]


def check_state_factory(state):
    "Return a property that checks if the current state is the desired state"

    @property
    def is_in_state(self):
        warnings.warn(
            "Using `machine.is_<state>` is deprecated. Use `machine.<state>.is_active` instead.",
            DeprecationWarning,
        )
        return bool(self.current_state == state)

    return is_in_state



"""
Traffic light machine
---------------------

Demonstrates the concept of ``cycle`` states.

"""
from statemachine import State
from statemachine import StateMachine


class TrafficLightMachine(StateMachine):
    "A traffic light machine"
    green = State("Green", initial=True)
    yellow = State("Yellow")
    red = State("Red")

    cycle = green.to(yellow) | yellow.to(red) | red.to(green)

    def on_cycle(self, event_data=None):
        return "Running {} from {} to {}".format(
            event_data.event,
            event_data.source.id,
            event_data.target.id,
        )



# coding: utf-8
import pytest

from statemachine.mixins import MachineMixin


class MyModel(MachineMixin):
    state_machine_name = "tests.conftest.CampaignMachine"

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        super(MyModel, self).__init__()

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.__dict__)


def test_mixin_should_instantiate_a_machine(campaign_machine):
    model = MyModel(state="draft")
    assert isinstance(model.statemachine, campaign_machine)
    assert model.state == "draft"
    assert model.statemachine.current_state == model.statemachine.draft


def test_mixin_should_raise_exception_if_machine_class_does_not_exist():
    class MyModelWithoutMachineName(MachineMixin):
        pass

    with pytest.raises(ValueError):
        MyModelWithoutMachineName()



import warnings
from collections import OrderedDict
from uuid import uuid4

from . import registry
from .event import Event
from .event import trigger_event_factory
from .exceptions import InvalidDefinition
from .graph import visit_connected_states
from .state import State
from .transition import Transition
from .transition_list import TransitionList
from .utils import check_state_factory
from .utils import qualname
from .utils import ugettext as _


class StateMachineMetaclass(type):
    def __init__(cls, name, bases, attrs):
        super(StateMachineMetaclass, cls).__init__(name, bases, attrs)
        registry.register(cls)
        cls._abstract = True
        cls.name = cls.__name__
        cls.states = []
        cls._events = OrderedDict()
        cls.states_map = {}
        cls.add_inherited(bases)
        cls.add_from_attributes(attrs)

        for state in cls.states:
            setattr(cls, "is_{}".format(state.id), check_state_factory(state))

        cls._set_special_states()
        cls._check()

    def _set_special_states(cls):
        if not cls.states:
            return
        initials = [s for s in cls.states if s.initial]
        if len(initials) != 1:
            raise InvalidDefinition(
                _(
                    "There should be one and only one initial state. "
                    "Your currently have these: {!r}".format(initials)
                )
            )
        cls.initial_state = initials[0]
        cls.final_states = [state for state in cls.states if state.final]

    def _disconnected_states(cls, starting_state):
        visitable_states = set(visit_connected_states(starting_state))
        return set(cls.states) - visitable_states

    def _check(cls):

        # do not validate the base class
        name = qualname(cls)
        if name in (
            "statemachine.factory_2.StateMachine",
            "statemachine.factory_3.StateMachine",
        ):
            return

        cls._abstract = False

        if not cls.states:
            raise InvalidDefinition(_("There are no states."))

        if not cls._events:
            raise InvalidDefinition(_("There are no events."))

        disconnected_states = cls._disconnected_states(cls.initial_state)
        if disconnected_states:
            raise InvalidDefinition(
                _(
                    "There are unreachable states. "
                    "The statemachine graph should have a single component. "
                    "Disconnected states: [{}]".format(disconnected_states)
                )
            )

        final_state_with_invalid_transitions = [
            state for state in cls.final_states if state.transitions
        ]

        if final_state_with_invalid_transitions:
            raise InvalidDefinition(
                _(
                    "Cannot declare transitions from final state. Invalid state(s): {}".format(
                        [s.id for s in final_state_with_invalid_transitions]
                    )
                )
            )

    def add_inherited(cls, bases):
        for base in bases:
            for state in getattr(base, "states", []):
                cls.add_state(state.id, state)

            events = getattr(base, "_events", {})
            for event in events.values():
                cls.add_event(event.name)

    def add_from_attributes(cls, attrs):
        for key, value in sorted(attrs.items(), key=lambda pair: pair[0]):
            if isinstance(value, State):
                cls.add_state(key, value)
            elif isinstance(value, (Transition, TransitionList)):
                cls.add_event(key, value)
            elif getattr(value, "_callbacks_to_update", None):
                cls._add_unbounded_callback(key, value)

    def _add_unbounded_callback(cls, attr_name, func):
        if func._is_event:
            # if func is an event, the `attr_name` will be replaced by an event trigger,
            # so we'll also give the ``func`` a new unique name to be used by the callback
            # machinery.
            cls.add_event(attr_name, func._transitions)
            attr_name = "_{}_{}".format(attr_name, uuid4().hex)
            setattr(cls, attr_name, func)

        for ref in func._callbacks_to_update:
            ref(attr_name)

    def add_state(cls, id, state):
        state._set_id(id)
        cls.states.append(state)
        cls.states_map[state.value] = state

        # also register all events associated directly with transitions
        for event in state.transitions.unique_events:
            cls.add_event(event)

    def add_event(cls, event, transitions=None):
        if transitions is not None:
            transitions.add_event(event)

        if event not in cls._events:
            event_instance = Event(event)
            cls._events[event] = event_instance
            event_trigger = trigger_event_factory(event)
            setattr(cls, event, event_trigger)

        return cls._events[event]

    @property
    def transitions(self):
        warnings.warn(
            "Class level property `transitions` is deprecated. Use `events` instead.",
            DeprecationWarning,
        )
        return self.events

    @property
    def events(self):
        return list(self._events.values())



from datetime import datetime

import pytest


@pytest.fixture
def current_time():
    return datetime.now()


@pytest.fixture
def campaign_machine():
    "Define a new class for each test"
    from statemachine import State, StateMachine

    class CampaignMachine(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True)
        producing = State("Being produced")
        closed = State("Closed")

        add_job = draft.to(draft) | producing.to(producing)
        produce = draft.to(producing)
        deliver = producing.to(closed)

    return CampaignMachine


@pytest.fixture
def campaign_machine_with_validator():
    "Define a new class for each test"
    from statemachine import State, StateMachine

    class CampaignMachine(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True)
        producing = State("Being produced")
        closed = State("Closed")

        add_job = draft.to(draft) | producing.to(producing)
        produce = draft.to(producing, validators="can_produce")
        deliver = producing.to(closed)

        def can_produce(*args, **kwargs):
            if "goods" not in kwargs:
                raise LookupError("Goods not found.")

    return CampaignMachine


@pytest.fixture
def campaign_machine_with_final_state():
    "Define a new class for each test"
    from statemachine import State, StateMachine

    class CampaignMachine(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True)
        producing = State("Being produced")
        closed = State("Closed", final=True)

        add_job = draft.to(draft) | producing.to(producing)
        produce = draft.to(producing)
        deliver = producing.to(closed)

    return CampaignMachine


@pytest.fixture
def campaign_machine_with_values():
    "Define a new class for each test"
    from statemachine import State, StateMachine

    class CampaignMachineWithKeys(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True, value=1)
        producing = State("Being produced", value=2)
        closed = State("Closed", value=3)

        add_job = draft.to(draft) | producing.to(producing)
        produce = draft.to(producing)
        deliver = producing.to(closed)

    return CampaignMachineWithKeys


@pytest.fixture
def traffic_light_machine():
    from tests.examples.traffic_light_machine import TrafficLightMachine

    return TrafficLightMachine


@pytest.fixture
def OrderControl():
    from tests.examples.order_control_machine import OrderControl

    return OrderControl


@pytest.fixture
def AllActionsMachine():
    from tests.examples.all_actions_machine import AllActionsMachine

    return AllActionsMachine


@pytest.fixture(autouse=True)
def add_machines_to_doctest(doctest_namespace, traffic_light_machine):
    doctest_namespace["TrafficLightMachine"] = traffic_light_machine


@pytest.fixture
def classic_traffic_light_machine():
    from statemachine import StateMachine, State

    class TrafficLightMachine(StateMachine):
        green = State("Green", initial=True)
        yellow = State("Yellow")
        red = State("Red")

        slowdown = green.to(yellow)
        stop = yellow.to(red)
        go = red.to(green)

    return TrafficLightMachine


@pytest.fixture
def reverse_traffic_light_machine():
    from statemachine import StateMachine, State

    class ReverseTrafficLightMachine(StateMachine):
        "A traffic light machine"
        green = State("Green", initial=True)
        yellow = State("Yellow")
        red = State("Red")

        stop = red.from_(yellow, green, red)
        cycle = (
            green.from_(red)
            | yellow.from_(green)
            | red.from_(yellow)
            | red.from_.itself()
        )

    return ReverseTrafficLightMachine


@pytest.fixture
def approval_machine(current_time):
    from statemachine import StateMachine, State

    class ApprovalMachine(StateMachine):
        "A workflow machine"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        rejected = State("Rejected")

        completed = State("Completed")

        validate = requested.to(accepted, cond="is_ok") | requested.to(rejected)

        @validate
        def do_validate(self, *args, **kwargs):
            if self.model.is_ok():
                self.model.accepted_at = current_time
                return self.model
            else:
                self.model.rejected_at = current_time
                return self.model

        @accepted.to(completed)
        def complete(self):
            self.model.completed_at = current_time

        @requested.to(requested)
        def update(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self.model, k, v)
            return self.model

        @rejected.to(requested)
        def retry(self):
            self.model.rejected_at = None
            return self.model

    return ApprovalMachine



# coding: utf-8
import warnings
from typing import Any
from typing import Optional
from typing import Text

from .callbacks import Callbacks
from .exceptions import StateMachineError
from .transition import Transition
from .transition_list import TransitionList
from .utils import ugettext as _


class State(object):
    """
    A State in a state machine describes a particular behaviour of the machine.
    When we say that a machine is “in” a state, it means that the machine behaves
    in the way that state describes.

    Args:
        name: An human readable representation of the state.
        value: A specific value to the storage and retrieval of states.
            If specified, you can use It to map a more friendly representation to a low-level
            value.
        initial: Set `' True`` if the ``State`` is the initial one. There must be one and only
            one initial state in a statemachine. Defaults to ``False``.
        final: Set ``True`` if represents a final state. A machine can have
            optionally many final states. Final states have no :ref:`transition` starting from It.
            Defaults to ``False``.
        enter: One or more callbacks assigned to be executed when the state is entered.
            See :ref:`actions`.
        exit: One or more callbacks assigned to be executed when the state is exited.
            See :ref:`actions`.

    State is a core component on how this library implements an expressive API to declare
    StateMachines.

    >>> from statemachine import State

    Given a few states...

    >>> draft = State("Draft", initial=True)

    >>> producing = State("Producing")

    >>> closed = State('Closed', final=True)

    Transitions are declared using the :func:`State.to` or :func:`State.from_` (reversed) methods.

    >>> draft.to(producing)
    TransitionList([Transition(State('Draft', ...

    The result is a `TransitionList`. Don't worry about this internal class. But the good
    thing is that it implements the ``OR`` operator to combine transitions, so you can use the
    ``|`` syntax to compound a list of transitions and assign to the same event.

    >>> transitions = draft.to(draft) | draft.to(producing)

    >>> [(t.source.name, t.target.name) for t in transitions]
    [('Draft', 'Draft'), ('Draft', 'Producing')]

    There are handy shortcuts that you can use to express this same set of transitions.

    The first one, ``draft.to(draft)``, is also called a :ref:`self-transition`, and can be
    expressed using an alternative syntax:

    >>> draft.to.itself()
    TransitionList([Transition(State('Draft', ...

    You can even pass a list of target states to declare at once all transitions starting
    from the same state.

    >>> transitions = draft.to(draft, producing, closed)

    >>> [(t.source.name, t.target.name) for t in transitions]
    [('Draft', 'Draft'), ('Draft', 'Producing'), ('Draft', 'Closed')]

    """

    def __init__(
        self, name, value=None, initial=False, final=False, enter=None, exit=None
    ):
        # type: (Text, Optional[Any], bool, bool, Optional[Any], Optional[Any]) -> None
        self.name = name
        self.value = value
        self._id = None  # type: Optional[Text]
        self._initial = initial
        self.transitions = TransitionList()
        self._final = final
        self.enter = Callbacks().add(enter)
        self.exit = Callbacks().add(exit)

    def _setup(self, resolver):
        self.enter.setup(resolver)
        self.exit.setup(resolver)

    def _add_observer(self, *resolvers):
        for r in resolvers:
            self.enter.add(
                "on_enter_state", resolver=r, prepend=True, suppress_errors=True
            )
            self.enter.add(
                "on_enter_{}".format(self.id), resolver=r, suppress_errors=True
            )
            self.exit.add(
                "on_exit_state", resolver=r, prepend=True, suppress_errors=True
            )
            self.exit.add(
                "on_exit_{}".format(self.id), resolver=r, suppress_errors=True
            )

    def __repr__(self):
        return "{}({!r}, id={!r}, value={!r}, initial={!r}, final={!r})".format(
            type(self).__name__,
            self.name,
            self.id,
            self.value,
            self.initial,
            self.final,
        )

    def __get__(self, machine, owner):
        self.machine = machine
        return self

    def __set__(self, instance, value):
        raise StateMachineError(
            _("State overriding is not allowed. Trying to add '{}' to {}").format(
                value, self.id
            )
        )

    @property
    def id(self):
        return self._id

    @property
    def identifier(self):
        warnings.warn(
            "`State.identifier` is deprecated. Use `State.id` instead.",
            DeprecationWarning,
        )
        return self.id

    def _set_id(self, id):
        self._id = id
        if self.value is None:
            self.value = id

    def _to_(self, *states, **kwargs):
        transitions = TransitionList(
            Transition(self, state, **kwargs) for state in states
        )
        self.transitions.add_transitions(transitions)
        return transitions

    def _from_(self, *states, **kwargs):
        transitions = TransitionList()
        for origin in states:
            transition = Transition(origin, self, **kwargs)
            origin.transitions.add_transitions(transition)
            transitions.add_transitions(transition)
        return transitions

    def _get_proxy_method_to_itself(self, method):
        def proxy(*states, **kwargs):
            return method(*states, **kwargs)

        def proxy_to_itself(**kwargs):
            return proxy(self, **kwargs)

        proxy.itself = proxy_to_itself
        return proxy

    @property
    def to(self):
        """Create transitions to the given target states."""
        return self._get_proxy_method_to_itself(self._to_)

    @property
    def from_(self):
        """Create transitions from the given target states (reversed)."""
        return self._get_proxy_method_to_itself(self._from_)

    @property
    def initial(self):
        return self._initial

    @property
    def final(self):
        return self._final

    @property
    def is_active(self):
        return self.machine.current_state == self



# coding: utf-8
import pytest

from statemachine import exceptions
from statemachine import State
from statemachine import StateMachine


class MyModel(object):
    "A class that can be used to hold arbitrary key/value pairs as attributes."

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.__dict__)


def test_machine_repr(campaign_machine):
    model = MyModel()
    machine = campaign_machine(model)
    assert (
        repr(machine) == "CampaignMachine(model=MyModel({'state': 'draft'}), "
        "state_field='state', current_state='draft')"
    )


def test_machine_should_be_at_start_state(campaign_machine):
    model = MyModel()
    machine = campaign_machine(model)

    assert [s.value for s in campaign_machine.states] == [
        "closed",
        "draft",
        "producing",
    ]
    assert [t.name for t in campaign_machine.events] == [
        "add_job",
        "deliver",
        "produce",
    ]  # noqa: E501

    assert model.state == "draft"
    assert machine.current_state == machine.draft


def test_machine_should_only_allow_only_one_initial_state():
    with pytest.raises(exceptions.InvalidDefinition):

        class CampaignMachine(StateMachine):
            "A workflow machine"
            draft = State("Draft", initial=True)
            producing = State("Being produced")
            closed = State(
                "Closed", initial=True
            )  # Should raise an Exception right after the class is defined

            add_job = draft.to(draft) | producing.to(producing)
            produce = draft.to(producing)
            deliver = producing.to(closed)


def test_machine_should_not_allow_transitions_from_final_state():
    with pytest.raises(exceptions.InvalidDefinition):

        class CampaignMachine(StateMachine):
            "A workflow machine"
            draft = State("Draft", initial=True)
            producing = State("Being produced")
            closed = State("Closed", final=True)

            add_job = draft.to(draft) | producing.to(producing) | closed.to(draft)
            produce = draft.to(producing)
            deliver = producing.to(closed)


def test_should_change_state(campaign_machine):
    model = MyModel()
    machine = campaign_machine(model)

    assert model.state == "draft"
    assert machine.current_state == machine.draft

    machine.produce()

    assert model.state == "producing"
    assert machine.current_state == machine.producing


def test_should_run_a_transition_that_keeps_the_state(campaign_machine):
    model = MyModel()
    machine = campaign_machine(model)

    assert model.state == "draft"
    assert machine.current_state == machine.draft

    machine.add_job()
    assert model.state == "draft"
    assert machine.current_state == machine.draft

    machine.produce()
    assert model.state == "producing"
    assert machine.current_state == machine.producing

    machine.add_job()
    assert model.state == "producing"
    assert machine.current_state == machine.producing


def test_should_change_state_with_multiple_machine_instances(campaign_machine):
    model1 = MyModel()
    model2 = MyModel()
    machine1 = campaign_machine(model1)
    machine2 = campaign_machine(model2)

    assert machine1.current_state == campaign_machine.draft
    assert machine2.current_state == campaign_machine.draft

    p1 = machine1.produce
    p2 = machine2.produce

    p2()
    assert machine1.current_state == campaign_machine.draft
    assert machine2.current_state == campaign_machine.producing

    p1()
    assert machine1.current_state == campaign_machine.producing
    assert machine2.current_state == campaign_machine.producing


@pytest.mark.parametrize(
    "current_state, transition",
    [
        ("draft", "deliver"),
        ("closed", "add_job"),
    ],
)
def test_call_to_transition_that_is_not_in_the_current_state_should_raise_exception(
    campaign_machine, current_state, transition
):

    model = MyModel(state=current_state)
    machine = campaign_machine(model)

    assert machine.current_state.value == current_state

    with pytest.raises(exceptions.TransitionNotAllowed):
        machine.send(transition)


def test_machine_should_list_allowed_events_in_the_current_state(campaign_machine):

    model = MyModel()
    machine = campaign_machine(model)

    assert model.state == "draft"
    assert [t.name for t in machine.allowed_events] == ["add_job", "produce"]

    machine.produce()
    assert model.state == "producing"
    assert [t.name for t in machine.allowed_events] == ["add_job", "deliver"]

    deliver = machine.allowed_events[1]

    deliver()
    assert model.state == "closed"
    assert machine.allowed_events == []


def test_machine_should_run_a_transition_by_his_key(campaign_machine):

    model = MyModel()
    machine = campaign_machine(model)

    assert model.state == "draft"

    machine.send("add_job")
    assert model.state == "draft"
    assert machine.current_state == machine.draft

    machine.send("produce")
    assert model.state == "producing"
    assert machine.current_state == machine.producing


def test_machine_should_raise_an_exception_if_a_transition_by_his_key_is_not_found(
    campaign_machine,
):
    model = MyModel()
    machine = campaign_machine(model)

    assert model.state == "draft"

    with pytest.raises(exceptions.TransitionNotAllowed):
        machine.send("go_horse")


def test_machine_should_use_and_model_attr_other_than_state(campaign_machine):
    model = MyModel(status="producing")
    machine = campaign_machine(model, state_field="status")

    assert getattr(model, "state", None) is None
    assert model.status == "producing"
    assert machine.current_state == machine.producing

    machine.deliver()

    assert model.status == "closed"
    assert machine.current_state == machine.closed


def test_cant_assign_an_invalid_state_directly(campaign_machine):
    machine = campaign_machine()
    with pytest.raises(exceptions.InvalidStateValue):
        machine.current_state_value = "non existing state"


def test_should_allow_validate_data_for_transition(campaign_machine_with_validator):
    model = MyModel()
    machine = campaign_machine_with_validator(model)

    with pytest.raises(LookupError):
        machine.produce()

    machine.produce(goods="something")

    assert model.state == "producing"


def test_should_check_if_is_in_status(campaign_machine):
    model = MyModel()
    machine = campaign_machine(model)

    assert machine.draft.is_active
    assert not machine.producing.is_active
    assert not machine.closed.is_active

    machine.produce()

    assert not machine.draft.is_active
    assert machine.producing.is_active
    assert not machine.closed.is_active

    machine.deliver()

    assert not machine.draft.is_active
    assert not machine.producing.is_active
    assert machine.closed.is_active


def test_defined_value_must_be_assigned_to_models(campaign_machine_with_values):
    model = MyModel()
    machine = campaign_machine_with_values(model)

    assert model.state == 1
    machine.produce()
    assert model.state == 2
    machine.deliver()
    assert model.state == 3


def test_state_machine_without_model(campaign_machine):
    machine = campaign_machine()
    assert machine.draft.is_active
    assert not machine.producing.is_active
    assert not machine.closed.is_active

    machine.produce()

    assert not machine.draft.is_active
    assert machine.producing.is_active
    assert not machine.closed.is_active


@pytest.mark.parametrize(
    "model, machine_name, start_value",
    [
        (None, "campaign_machine", "producing"),
        (None, "campaign_machine_with_values", 2),
        (MyModel(), "campaign_machine", "producing"),
        (MyModel(), "campaign_machine_with_values", 2),
    ],
)
def test_state_machine_with_a_start_value(request, model, machine_name, start_value):
    machine_cls = request.getfixturevalue(machine_name)
    machine = machine_cls(model, start_value=start_value)
    assert not machine.draft.is_active
    assert machine.producing.is_active
    assert not model or model.state == start_value


@pytest.mark.parametrize(
    "model, machine_name, start_value",
    [
        (None, "campaign_machine", "tapioca"),
        (None, "campaign_machine_with_values", 99),
        (MyModel(), "campaign_machine", "tapioca"),
        (MyModel(), "campaign_machine_with_values", 99),
    ],
)
def test_state_machine_with_a_invalid_start_value(
    request, model, machine_name, start_value
):
    machine_cls = request.getfixturevalue(machine_name)
    with pytest.raises(exceptions.InvalidStateValue):
        machine_cls(model, start_value=start_value)


def test_should_not_create_instance_of_machine_without_states():

    with pytest.raises(exceptions.InvalidDefinition):

        class EmptyMachine(StateMachine):
            "An empty machine"
            pass


def test_should_not_create_instance_of_machine_without_transitions():

    with pytest.raises(exceptions.InvalidDefinition):

        class NoTransitionsMachine(StateMachine):
            "A machine without transitions"
            initial = State("initial", initial=True)


def test_should_not_create_disconnected_machine():

    with pytest.raises(exceptions.InvalidDefinition) as e:

        class BrokenTrafficLightMachine(StateMachine):
            "A broken traffic light machine"
            green = State("Green", initial=True)
            yellow = State("Yellow")
            blue = State("Blue")  # This state is unreachable

            cycle = green.to(yellow) | yellow.to(green)

        assert "Blue" in e.message
        assert "Green" not in e.message


def test_should_not_create_big_disconnected_machine():
    with pytest.raises(exceptions.InvalidDefinition) as e:

        class BrokenTrafficLightMachine(StateMachine):
            "A broken traffic light machine"
            green = State("Green", initial=True)
            yellow = State("Yellow")
            magenta = State("Magenta")  # This state is unreachable
            red = State("Red")
            cyan = State("Cyan")
            blue = State("Blue")  # This state is also unreachable

            cycle = green.to(yellow)
            diverge = green.to(cyan) | cyan.to(red)
            validate = yellow.to(green)

        assert "Magenta" in e.message
        assert "Blue" in e.message
        assert "Cyan" not in e.message


def test_state_value_is_correct():
    STATE_NEW = 0
    STATE_DRAFT = 1

    class ValueTestModel(StateMachine):
        new = State(STATE_NEW, value=STATE_NEW, initial=True)
        draft = State(STATE_DRAFT, value=STATE_DRAFT)

        write = new.to(draft)

    model = ValueTestModel()
    assert model.new.value == STATE_NEW
    assert model.draft.value == STATE_DRAFT


def test_final_states(campaign_machine_with_final_state):
    model = MyModel()
    machine = campaign_machine_with_final_state(model)
    final_states = machine.final_states
    assert len(final_states) == 1
    assert final_states[0].name == "Closed"


def test_should_not_override_states_properties(campaign_machine):

    machine = campaign_machine()
    with pytest.raises(exceptions.StateMachineError) as e:
        machine.draft = "something else"

    assert (
        "State overriding is not allowed. Trying to add 'something else' to draft"
        in str(e)
    )



"""
All actions machine
===================

A StateMachine that exercices all possible :ref:`Actions` and :ref:`Guards`.

"""
import mock

from statemachine import State
from statemachine import StateMachine


class AllActionsMachine(StateMachine):

    initial = State("Initial", initial=True)
    final = State("Final", final=True)

    go = initial.to(
        final,
        validators=["validation_1", "validation_2"],
        cond=["condition_1", "condition_2"],
        unless=["unless_1", "unless_2"],
        on=["on_inline_1", "on_inline_2"],
        before=["before_go_inline_1", "before_go_inline_2"],
        after=["after_go_inline_1", "after_go_inline_2"],
    )

    def __init__(self, *args, **kwargs):
        self.spy = mock.Mock(side_effect=lambda x: x)
        super(AllActionsMachine, self).__init__(*args, **kwargs)

    # validators and guards

    def validation_1(self):
        # this method may raise an exception
        return self.spy("validation_1")

    def validation_2(self):
        # this method may raise an exception
        return self.spy("validation_2")

    def condition_1(self):
        self.spy("condition_1")
        return True

    def condition_2(self):
        self.spy("condition_2")
        return True

    def unless_1(self):
        self.spy("unless_1")
        return False

    def unless_2(self):
        self.spy("unless_2")
        return False

    # generics state

    def on_enter_state(self):
        return self.spy("on_enter_state")

    def on_exit_state(self):
        return self.spy("on_exit_state")

    # generics transition

    def before_transition(self):
        return self.spy("before_transition")

    def on_transition(self):
        return self.spy("on_transition")

    def after_transition(self):
        return self.spy("after_transition")

    # before / after specific

    @go.before
    def before_go_decor(self):
        return self.spy("before_go_decor")

    def before_go_inline_1(self):
        return self.spy("before_go_inline_1")

    def before_go_inline_2(self):
        return self.spy("before_go_inline_2")

    def before_go(self):
        return self.spy("before_go")

    @go.on
    def go_on_decor(self):
        return self.spy("go_on_decor")

    def on_inline_1(self):
        return self.spy("on_inline_1")

    def on_inline_2(self):
        return self.spy("on_inline_2")

    def on_go(self):
        return self.spy("on_go")

    @go.after
    def after_go_decor(self):
        return self.spy("after_go_decor")

    def after_go_inline_1(self):
        return self.spy("after_go_inline_1")

    def after_go_inline_2(self):
        return self.spy("after_go_inline_2")

    def after_go(self):
        return self.spy("after_go")

    # enter / exit specific

    @initial.enter
    def enter_initial_decor(self):
        return self.spy("enter_initial_decor")

    def on_enter_initial(self):
        return self.spy("on_enter_initial")

    @initial.exit
    def exit_initial_decor(self):
        return self.spy("exit_initial_decor")

    def on_exit_initial(self):
        return self.spy("on_exit_initial")

    def on_enter_final(self):
        return self.spy("on_enter_final")

    def on_exit_final(self):
        "hopefully this will not be called"
        return self.spy("on_exit_final")


# %%
# Testing
# -------

machine = AllActionsMachine()
spy = machine.spy


# %%
# Only before/on actions have their result collected.

result = machine.go()
assert result == [
    "before_transition",
    "before_go_inline_1",
    "before_go_inline_2",
    "before_go_decor",
    "before_go",
    "on_transition",
    "on_inline_1",
    "on_inline_2",
    "go_on_decor",
    "on_go",
]

# %%
# Checking the method resolution order

assert spy.call_args_list == [
    mock.call("on_enter_state"),
    mock.call("enter_initial_decor"),
    mock.call("on_enter_initial"),
    mock.call("validation_1"),
    mock.call("validation_2"),
    mock.call("condition_1"),
    mock.call("condition_2"),
    mock.call("unless_1"),
    mock.call("unless_2"),
    mock.call("before_transition"),
    mock.call("before_go_inline_1"),
    mock.call("before_go_inline_2"),
    mock.call("before_go_decor"),
    mock.call("before_go"),
    mock.call("on_exit_state"),
    mock.call("exit_initial_decor"),
    mock.call("on_exit_initial"),
    mock.call("on_transition"),
    mock.call("on_inline_1"),
    mock.call("on_inline_2"),
    mock.call("go_on_decor"),
    mock.call("on_go"),
    mock.call("on_enter_state"),
    mock.call("on_enter_final"),
    mock.call("after_go_inline_1"),
    mock.call("after_go_inline_2"),
    mock.call("after_go_decor"),
    mock.call("after_go"),
    mock.call("after_transition"),
]



# coding: utf-8
from .exceptions import AttrNotFound
from .exceptions import InvalidDefinition
from .utils import ensure_iterable
from .utils import ugettext as _


class CallbackWrapper(object):
    """A thin wrapper that ensures the targef callback is a proper callable.

    At first, `func` can be a string or a callable, and even if it's already
    a callable, his signature can mismatch.

    After instantiation, `.setup(resolver)` must be called before any real
    call is performed, to allow the proper callback resolution.
    """

    def __init__(self, func, suppress_errors=False):
        self.func = func
        self.suppress_errors = suppress_errors
        self._callback = None

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.func)

    def __eq__(self, other):
        return self.func == getattr(other, "func", other)

    def __hash__(self):
        return id(self)

    def _update_func(self, func):
        self.func = func

    def setup(self, resolver):
        """
        Resolves the `func` into a usable callable.

        Args:
            resolver (callable): A method responsible to build and return a valid callable that
                can receive arbitrary parameters like `*args, **kwargs`.
        """
        try:
            self._callback = resolver(self.func)
            return True
        except AttrNotFound:
            if not self.suppress_errors:
                raise
            return False

    def __call__(self, *args, **kwargs):
        if self._callback is None:
            raise InvalidDefinition(
                _("Callback {!r} not property configured.").format(self)
            )
        return self._callback(*args, **kwargs)


class ConditionWrapper(CallbackWrapper):
    def __init__(self, func, suppress_errors=False, expected_value=True):
        super(ConditionWrapper, self).__init__(func, suppress_errors)
        self.expected_value = expected_value

    def __call__(self, *args, **kwargs):
        return (
            super(ConditionWrapper, self).__call__(*args, **kwargs)
            == self.expected_value
        )


class Callbacks(object):
    def __init__(self, resolver=None, factory=CallbackWrapper):
        self.items = []
        self._resolver = resolver
        self.factory = factory

    def __repr__(self):
        return "{}({!r}, factory={!r})".format(
            type(self).__name__, self.items, self.factory
        )

    def setup(self, resolver):
        """Validate configuracions"""
        self._resolver = resolver
        self.items = [
            callback for callback in self.items if callback.setup(self._resolver)
        ]

    def _add_unbounded_callback(self, func, is_event=False, transitions=None):
        """This list was a targed for adding a func using decorator
        `@<state|event>[.on|before|after|enter|exit]` syntax.

        If we assign ``func`` directly as callable on the ``items`` list,
        this will result in an `unbounded method error`, with `func` expecting a parameter
        ``self`` not defined.

        The implemented solution is to resolve the colision giving the func a reference method.
        To update It's callback when the name is resolved on the
        :func:`StateMachineMetaclass.add_from_attributes`.
        If the ``func`` is bounded It will be used directly, if not, it's ref will be replaced
        by the given attr name and on `statemachine._setup()` the dynamic name will be resolved
        properly.

        Args:
            func (callable): The decorated method to add on the transitions occurs.
            is_event (bool): If the func is also an event, we'll create a trigger and link the
                event name to the transitions.
            transitions (TransitionList): If ``is_event``, the transitions to be attached to the
                event.

        """
        callback = self._add(func)
        if not getattr(func, "_callbacks_to_update", None):
            func._callbacks_to_update = set()
        func._callbacks_to_update.add(callback._update_func)
        func._is_event = is_event
        func._transitions = transitions

        return func

    def __call__(self, callback):
        return self._add_unbounded_callback(callback)

    def __iter__(self):
        return iter(self.items)

    def clear(self):
        self.items = []

    def call(self, *args, **kwargs):
        return [callback(*args, **kwargs) for callback in self.items]

    def _add(self, func, resolver=None, prepend=False, **kwargs):
        if func in self.items:
            return

        resolver = resolver or self._resolver

        callback = self.factory(func, **kwargs)
        if resolver is not None and not callback.setup(resolver):
            return

        if prepend:
            self.items.insert(0, callback)
        else:
            self.items.append(callback)
        return callback

    def add(self, callbacks, **kwargs):
        if callbacks is None:
            return self

        unprepared = ensure_iterable(callbacks)
        for func in unprepared:
            self._add(func, **kwargs)

        return self



from statemachine import State


def test_transition_list_or_operator():
    s1 = State("s1", initial=True)
    s2 = State("s2")
    s3 = State("s3")
    s4 = State("s4", final=True)

    t12 = s1.to(s2)
    t23 = s2.to(s3)
    t34 = s3.to(s4)

    cycle = t12 | t23 | t34

    assert [(t.source.name, t.target.name) for t in t12] == [("s1", "s2")]
    assert [(t.source.name, t.target.name) for t in t23] == [("s2", "s3")]
    assert [(t.source.name, t.target.name) for t in t34] == [("s3", "s4")]
    assert [(t.source.name, t.target.name) for t in cycle] == [
        ("s1", "s2"),
        ("s2", "s3"),
        ("s3", "s4"),
    ]



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest

from statemachine import exceptions
from statemachine import State
from statemachine import StateMachine


class Request(object):
    def __init__(self, state="requested"):
        self.state = None
        self._is_ok = False

    def is_ok(self):
        return self._is_ok


def test_transition_should_choose_final_state_on_multiple_possibilities(
    approval_machine, current_time
):
    # given
    model = Request()
    machine = approval_machine(model)

    model._is_ok = False

    # when
    assert machine.validate() == model

    # then
    assert model.rejected_at == current_time
    assert machine.rejected.is_active

    # given
    model._is_ok = True

    # when
    assert machine.retry() == model

    # then
    assert model.rejected_at is None
    assert machine.requested.is_active

    # when
    assert machine.validate() == model

    # then
    assert model.accepted_at == current_time
    assert machine.accepted.is_active


def test_transition_to_first_that_executes_if_multiple_targets():
    class ApprovalMachine(StateMachine):
        "A workflow"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        rejected = State("Rejected")

        validate = requested.to(accepted, rejected)

    machine = ApprovalMachine()

    machine.validate()
    assert machine.accepted.is_active


def test_do_not_transition_if_multiple_targets_with_guard():
    def never_will_pass(event_data):
        return False

    class ApprovalMachine(StateMachine):
        "A workflow"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        rejected = State("Rejected")

        validate = (
            requested.to(accepted, cond=never_will_pass)
            | requested.to(rejected, cond="also_never_will_pass")
            | requested.to(requested, cond="this_also_never_will_pass")
        )

        @property
        def also_never_will_pass(self):
            return False

        def this_also_never_will_pass(self, event_data):
            return False

    machine = ApprovalMachine()

    with pytest.raises(exceptions.TransitionNotAllowed):
        machine.validate()
    assert machine.requested.is_active


def test_check_invalid_reference_to_conditions():
    class ApprovalMachine(StateMachine):
        "A workflow"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        rejected = State("Rejected")

        validate = requested.to(accepted, cond="not_found_condition") | requested.to(
            rejected
        )

    with pytest.raises(exceptions.InvalidDefinition):
        ApprovalMachine()


def test_should_change_to_returned_state_on_multiple_target_with_combined_transitions():
    class ApprovalMachine(StateMachine):
        "A workflow"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        rejected = State("Rejected")
        completed = State("Completed")

        validate = (
            requested.to(accepted, cond="is_ok")
            | requested.to(rejected)
            | accepted.to(completed)
        )
        retry = rejected.to(requested)

        def on_validate(self):
            if self.accepted.is_active and self.model.is_ok():
                return "congrats!"

    # given
    model = Request()
    machine = ApprovalMachine(model)

    model._is_ok = False

    # when
    assert machine.validate() is None
    # then
    assert machine.rejected.is_active

    # given
    assert machine.retry() is None
    assert machine.requested.is_active
    model._is_ok = True

    # when
    assert machine.validate() is None
    # then
    assert machine.accepted.is_active

    # when
    assert machine.validate() == "congrats!"
    # then
    assert machine.completed.is_active

    with pytest.raises(exceptions.TransitionNotAllowed) as e:
        assert machine.validate()
        assert e.message == "Can't validate when in Completed."


def test_transition_on_execute_should_be_called_with_run_syntax(
    approval_machine, current_time
):
    # given
    model = Request()
    machine = approval_machine(model)

    model._is_ok = True

    # when
    assert machine.send("validate") == model
    # then
    assert model.accepted_at == current_time
    assert machine.accepted.is_active


def test_multiple_values_returned_with_multiple_targets():
    class ApprovalMachine(StateMachine):
        "A workflow"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        denied = State("Denied")

        @requested.to(accepted, denied)
        def validate(self):
            return 1, 2

    machine = ApprovalMachine()

    assert machine.validate() == (
        1,
        2,
    )


@pytest.mark.parametrize(
    "payment_failed, expected_state",
    [
        (False, "paid"),
        (True, "failed"),
    ],
)
def test_multiple_targets_using_or_starting_from_same_origin(
    payment_failed, expected_state
):
    class InvoiceStateMachine(StateMachine):
        unpaid = State("unpaid", initial=True)
        paid = State("paid")
        failed = State("failed")

        pay = (
            unpaid.to(paid, unless="payment_success")
            | failed.to(paid)
            | unpaid.to(failed)
        )

        def payment_success(self, event_data):
            return payment_failed

    invoice_fsm = InvoiceStateMachine()
    invoice_fsm.pay()
    assert invoice_fsm.current_state.id == expected_state


def test_order_control(OrderControl):
    control = OrderControl()
    assert control.add_to_order(3) == 3
    assert control.add_to_order(7) == 10

    control.receive_payment(4)
    with pytest.raises(exceptions.TransitionNotAllowed):
        control.process_order()

    control.receive_payment(6)
    control.process_order()

    control.ship_order()
    assert control.order_total == 10
    assert control.payments == [4, 6]
    assert control.completed.is_active



import importlib
import sys

import pydot

from ..statemachine import BaseStateMachine


class DotGraphMachine(object):
    graph_rankdir = "LR"
    """
    Diretion of the graph. Defaults to "LR" (option "TB" for top botton)
    http://www.graphviz.org/doc/info/attrs.html#d:rankdir
    """

    state_font_size = "10pt"
    """State font size"""

    state_active_penwidth = 2
    """Active state external line width"""

    state_active_fillcolor = "turquoise"

    transition_font_size = "9pt"
    """Transition font size"""

    def __init__(self, machine):
        self.machine = machine

    def _get_graph(self):
        machine = self.machine
        return pydot.Dot(
            "list",
            graph_type="digraph",
            label=machine.name,
            fontsize=self.state_font_size,
            rankdir=self.graph_rankdir,
        )

    def _initial_node(self):
        node = pydot.Node(
            "i",
            shape="circle",
            style="filled",
            fontsize="1pt",
            fixedsize="true",
            width=0.2,
            height=0.2,
        )
        node.set_fillcolor("black")
        return node

    def _initial_edge(self):
        return pydot.Edge(
            "i",
            self.machine.initial_state.id,
            label="",
            color="blue",
            fontsize=self.transition_font_size,
        )

    def _state_actions(self, state):
        entry = ", ".join(
            [getattr(action.func, "__name__", action.func) for action in state.enter]
        )
        exit = ", ".join(
            [getattr(action.func, "__name__", action.func) for action in state.exit]
        )

        if entry:
            entry = "entry / {}".format(entry)
        if exit:
            exit = "exit / {}".format(exit)

        actions = "\n".join(x for x in [entry, exit] if x)

        if actions:
            actions = "\n{}".format(actions)

        return actions

    def _state_as_node(self, state):
        actions = self._state_actions(state)

        node = pydot.Node(
            state.id,
            label="{}{}".format(state.name, actions),
            shape="rectangle",
            style="rounded, filled",
            fontsize=self.state_font_size,
            peripheries=2 if state.final else 1,
        )
        if state == self.machine.current_state:
            node.set_penwidth(self.state_active_penwidth)
            node.set_fillcolor(self.state_active_fillcolor)
        else:
            node.set_fillcolor("white")
        return node

    def _transition_as_edge(self, transition):
        def _get_condition_repr(cond):
            name = getattr(cond.func, "__name__", cond.func)
            return name if cond.expected_value else "!{}".format(name)

        cond = ", ".join([_get_condition_repr(cond) for cond in transition.cond])
        if cond:
            cond = "\n[{}]".format(cond)
        return pydot.Edge(
            transition.source.id,
            transition.target.id,
            label="{}{}".format(transition.event, cond),
            color="blue",
            fontsize=self.transition_font_size,
        )

    def get_graph(self):
        graph = self._get_graph()

        initial_node = self._initial_node()
        initial_edge = self._initial_edge()
        graph.add_node(initial_node)
        graph.add_edge(initial_edge)

        for state in self.machine.states:
            graph.add_node(self._state_as_node(state))
            for transition in state.transitions:
                graph.add_edge(self._transition_as_edge(transition))

        return graph

    def __call__(self):
        return self.get_graph()


def import_sm(qualname):
    module_name, class_name = qualname.rsplit(".", 1)
    module = importlib.import_module(module_name)
    smclass = getattr(module, class_name, None)
    if not smclass or not issubclass(smclass, BaseStateMachine):
        raise ValueError("{} is not a subclass of StateMachine".format(class_name))

    return smclass


def write_image(qualname, out):
    """
    Given a `qualname`, that is the fully qualified dotted path to a StateMachine
    classe, imports the class and generates a dot graph using the `pydot` lib.
    Writes the graph representation to the filename 'out' that will
    open/create and truncate such file and write on it a representation of
    the graph defined by the statemachine, in the format specified by
    the extension contained in the out path (out.ext).
    """
    smclass = import_sm(qualname)

    graph = DotGraphMachine(smclass).get_graph()
    out_extension = out.rsplit(".", 1)[1]
    graph.write(out, format=out_extension)


def main(argv=None):
    import argparse

    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION] <classpath> <out>",
        description="Generate diagrams for StateMachine classes.",
    )
    parser.add_argument(
        "classpath", help="A fully-qualified dotted path to the StateMachine class."
    )
    parser.add_argument(
        "out",
        help="File to generate the image using extension as the output format.",
    )

    args = parser.parse_args(argv)
    write_image(qualname=args.classpath, out=args.out)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from typing import Text

try:
    from django.utils.translation import ugettext as _
except Exception:

    def _(text):
        return text


from . import registry


class MachineMixin(object):
    """This mixing allows a model to automatically instantiate and assign an
    ``StateMachine``.
    """

    state_field_name = "state"  # type: Text
    """The model's state field name that will hold the state value."""

    state_machine_name = None  # type: Text
    """A fully qualified name of the class, where it can be imported."""

    state_machine_attr = "statemachine"  # type: Text
    """Name of the model's attribute that will hold the machine instance."""

    def __init__(self, *args, **kwargs):
        super(MachineMixin, self).__init__(*args, **kwargs)
        if not self.state_machine_name:
            raise ValueError(
                _("{!r} is not a valid state machine name.").format(
                    self.state_machine_name
                )
            )
        machine_cls = registry.get_machine_cls(self.state_machine_name)
        setattr(
            self,
            self.state_machine_attr,
            machine_cls(self, state_field=self.state_field_name),
        )



"""
Order control machine
---------------------

An StateMachine that demonstrates :ref:`Guards` being used to control the state flow.

"""
from statemachine import State
from statemachine import StateMachine


class OrderControl(StateMachine):
    waiting_for_payment = State("Waiting for payment", initial=True)
    processing = State("Processing")
    shipping = State("Shipping")
    completed = State("Completed", final=True)

    add_to_order = waiting_for_payment.to(waiting_for_payment)
    receive_payment = waiting_for_payment.to(
        processing, cond="payments_enough"
    ) | waiting_for_payment.to(waiting_for_payment, unless="payments_enough")
    process_order = processing.to(shipping, cond="payment_received")
    ship_order = shipping.to(completed)

    def __init__(self):
        self.order_total = 0
        self.payments = []
        self.payment_received = False
        super(OrderControl, self).__init__()

    def payments_enough(self, amount):
        return sum(self.payments) + amount >= self.order_total

    def before_add_to_order(self, amount):
        self.order_total += amount
        return self.order_total

    def before_receive_payment(self, amount):
        self.payments.append(amount)
        return self.payments

    def after_receive_payment(self):
        self.payment_received = True

    def on_enter_waiting_for_payment(self):
        self.payment_received = False



# coding: utf-8
import pytest

from statemachine.dispatcher import ensure_callable
from statemachine.dispatcher import ObjectConfig
from statemachine.dispatcher import resolver_factory


class Person(object):
    def __init__(self, first_name, last_name, legal_document=None):
        self.first_name = first_name
        self.last_name = last_name
        self.legal_document = legal_document

    def get_full_name(self):
        return "{} {}".format(self.first_name, self.last_name)


class Organization(object):
    def __init__(self, name, legal_document):
        self.name = name
        self.legal_document = legal_document

    def get_full_name(self):
        return self.name


class TestEnsureCallable:
    @pytest.fixture(
        params=[
            pytest.param([], id="no-args"),
            pytest.param([24, True, "Go!"], id="with-args"),
        ]
    )
    def args(self, request):
        return request.param

    @pytest.fixture(
        params=[
            pytest.param(dict(), id="no-kwargs"),
            pytest.param(dict(a="x", b="y"), id="with-kwargs"),
        ]
    )
    def kwargs(self, request):
        return request.param

    def test_return_same_object_if_already_a_callable(self):
        model = Person("Frodo", "Bolseiro")
        expected = model.get_full_name
        actual = ensure_callable(expected)
        assert actual.__name__ == expected.__name__
        assert actual.__doc__ == expected.__doc__

    def test_retrieve_a_method_from_its_name(self, args, kwargs):
        model = Person("Frodo", "Bolseiro")
        expected = model.get_full_name
        method = ensure_callable("get_full_name", model)

        assert method.__name__ == expected.__name__
        assert method.__doc__ == expected.__doc__
        assert method(*args, **kwargs) == "Frodo Bolseiro"

    def test_retrieve_a_callable_from_a_property_name(self, args, kwargs):
        model = Person("Frodo", "Bolseiro")
        method = ensure_callable("first_name", model)

        assert method(*args, **kwargs) == "Frodo"

    def test_retrieve_callable_from_a_property_name_that_should_keep_reference(
        self, args, kwargs
    ):
        model = Person("Frodo", "Bolseiro")
        method = ensure_callable("first_name", model)

        model.first_name = "Bilbo"

        assert method(*args, **kwargs) == "Bilbo"


class TestResolverFactory:
    @pytest.mark.parametrize(
        "attr, expected_value",
        [
            ("first_name", "Frodo"),
            ("last_name", "Bolseiro"),
            ("legal_document", "cnpj"),
            ("get_full_name", "The Lord fo the Rings"),
        ],
    )
    def test_should_chain_resolutions(self, attr, expected_value):
        person = Person("Frodo", "Bolseiro", "cpf")
        org = Organization("The Lord fo the Rings", "cnpj")

        resolver = resolver_factory(org, person)
        resolved_method = resolver(attr)
        assert resolved_method() == expected_value

    @pytest.mark.parametrize(
        "attr, expected_value",
        [
            ("first_name", "Frodo"),
            ("last_name", "Bolseiro"),
            ("legal_document", "cnpj"),
            ("get_full_name", "Frodo Bolseiro"),
        ],
    )
    def test_should_ignore_list_of_attrs(self, attr, expected_value):
        person = Person("Frodo", "Bolseiro", "cpf")
        org = Organization("The Lord fo the Rings", "cnpj")

        org_config = ObjectConfig(org, {"get_full_name"})

        resolver = resolver_factory(org_config, person)
        resolved_method = resolver(attr)
        assert resolved_method() == expected_value



from collections import deque


def visit_connected_states(state):
    visit = deque()
    already_visited = set()
    visit.append(state)
    while visit:
        state = visit.popleft()
        if state in already_visited:
            continue
        already_visited.add(state)
        yield state
        visit.extend(t.target for t in state.transitions)



import importlib
import re

from statemachine.contrib.diagram import DotGraphMachine
from statemachine.factory import StateMachineMetaclass


class MachineScraper(object):
    """Scrapes images of the statemachines defined into the examples for the gallery"""

    re_machine_module_name = re.compile(r"python-statemachine/(.*).py$")
    re_replace_png_extension = re.compile(r"\.png$")

    def __init__(self):
        self.seen = set()

    def __repr__(self):
        return "MachineScraper"

    def __call__(self, block, block_vars, gallery_conf):
        # Find all PNG files in the directory of this example.
        from sphinx_gallery.scrapers import figure_rst

        src_file = block_vars["src_file"]

        module_name = self.re_machine_module_name.findall(src_file)
        if len(module_name) != 1:
            return ""

        module_name = module_name[0].replace("/", ".")
        module = importlib.import_module(module_name)

        image_names = []
        image_path_iterator = block_vars["image_path_iterator"]
        for key, value in module.__dict__.items():
            if key.startswith("__"):
                continue

            if not isinstance(value, StateMachineMetaclass) or value._abstract:
                continue

            unique_key = "{}.{}".format(module_name, key)

            if unique_key in self.seen:
                continue
            self.seen.add(unique_key)

            image_path = image_path_iterator.next()
            image_path = self.re_replace_png_extension.sub(".svg", image_path)
            image_names.append(image_path)

            svg = DotGraphMachine(value).get_graph().create_svg().decode()
            with open(image_path, "w") as f:
                f.write(svg)

        # Use the `figure_rst` helper function to generate rST for image files
        return figure_rst(image_names, gallery_conf["src_dir"])



from statemachine import State
from statemachine import StateMachine


def test_assign_events_on_transitions():
    class TrafficLightMachine(StateMachine):
        "A traffic light machine"
        green = State("Green", initial=True)
        yellow = State("Yellow")
        red = State("Red")

        green.to(yellow, event="cycle slowdown slowdown")
        yellow.to(red, event="cycle stop")
        red.to(green, event="cycle go")

        def on_cycle(self, event_data=None):
            return "Running {} from {} to {}".format(
                event_data.event,
                event_data.transition.source.id,
                event_data.transition.target.id,
            )

    sm = TrafficLightMachine()

    assert sm.send("cycle") == "Running cycle from green to yellow"
    assert sm.send("cycle") == "Running cycle from yellow to red"
    assert sm.send("cycle") == "Running cycle from red to green"



# coding: utf-8
import mock
import pytest

from .models import MyModel
from statemachine import State
from statemachine import StateMachine


@pytest.fixture
def event_mock():
    return mock.MagicMock()


@pytest.fixture
def state_machine(event_mock):
    class CampaignMachine(StateMachine):
        draft = State("Draft", initial=True)
        producing = State("Being produced")
        closed = State("Closed")

        add_job = draft.to(draft) | producing.to(producing)
        produce = draft.to(producing)
        deliver = producing.to(closed)

        def on_enter_producing(self, param1=None, param2=None):
            event_mock.on_enter_producing(param1=param1, param2=param2)

        def on_exit_draft(self, param1=None, param2=None):
            event_mock.on_exit_draft(param1=param1, param2=param2)

        def on_enter_closed(self):
            event_mock.on_enter_closed()

        def on_exit_producing(self):
            event_mock.on_exit_producing()

    return CampaignMachine


def test_run_transition_pass_arguments_to_sub_transitions(
    state_machine,
    event_mock,
):
    model = MyModel(state="draft")
    machine = state_machine(model)

    machine.send("produce", param1="value1", param2="value2")

    assert model.state == "producing"
    event_mock.on_enter_producing.assert_called_with(param1="value1", param2="value2")
    event_mock.on_exit_draft.assert_called_with(param1="value1", param2="value2")

    machine.send("deliver", param3="value3")

    event_mock.on_enter_closed.assert_called_with()
    event_mock.on_exit_producing.assert_called_with()



# coding: utf-8
import warnings
from functools import partial

from .callbacks import Callbacks
from .callbacks import ConditionWrapper
from .events import Events


class Transition(object):
    """A transition holds reference to the source and target state.

    Args:
        source (State): The origin state of the transition.
        target (State): The target state of the transition.
        event (Optional[Union[str, List[str]]]): List of designators of events that trigger this
            transition. Can be either a list of strings, or a space-separated string list of event
            descriptors.
        validators (Optional[Union[str, Callable, List[Callable]]]): The validation callbacks to
            be invoked before the transition is executed.
        cond (Optional[Union[str, Callable, List[Callable]]]): The condition callbacks to be
            invoked before the transition is executed that should evaluate to `True`.
        unless (Optional[Union[str, Callable, List[Callable]]]): The condition callbacks to be
            invoked if the `cond` is False before the transition is executed.
        on (Optional[Union[str, Callable, List[Callable]]]): The callbacks to be invoked
            when the transition is executed.
        before (Optional[Union[str, Callable, List[Callable]]]): The callbacks to be invoked
            before the transition is executed.
        after (Optional[Union[str, Callable, List[Callable]]]): The callbacks to be invoked
            after the transition is executed.
    """

    def __init__(
        self,
        source,
        target,
        event=None,
        validators=None,
        cond=None,
        unless=None,
        on=None,
        before=None,
        after=None,
    ):

        self.source = source
        self.target = target
        self._events = Events().add(event)
        self.validators = Callbacks().add(validators)
        self.before = Callbacks().add(before)
        self.on = Callbacks().add(on)
        self.after = Callbacks().add(after)
        self.cond = (
            Callbacks(factory=ConditionWrapper)
            .add(cond)
            .add(unless, expected_value=False)
        )

    def __repr__(self):
        return "{}({!r}, {!r}, event={!r})".format(
            type(self).__name__, self.source, self.target, self.event
        )

    def _setup(self, resolver):
        self.validators.setup(resolver)
        self.cond.setup(resolver)
        self.before.setup(resolver)
        self.on.setup(resolver)
        self.after.setup(resolver)

    def _add_observer(self, *resolvers):
        for r in resolvers:
            before = partial(self.before.add, resolver=r, suppress_errors=True)
            on = partial(self.on.add, resolver=r, suppress_errors=True)
            after = partial(self.after.add, resolver=r, suppress_errors=True)

            before("before_transition", prepend=True)
            on("on_transition", prepend=True)

            for event in self._events:
                before("before_{}".format(event))
                on("on_{}".format(event))
                after("after_{}".format(event))

            after("after_transition")

    def _eval_cond(self, event_data):
        return all(
            condition(*event_data.args, **event_data.extended_kwargs)
            for condition in self.cond
        )

    def match(self, event):
        return self._events.match(event)

    @property
    def identifier(self):
        warnings.warn(
            "identifier is deprecated. Use `event` instead", DeprecationWarning
        )
        return self.event

    @property
    def event(self):
        return str(self._events)

    @property
    def events(self):
        return self._events

    def add_event(self, value):
        self._events.add(value)

    def execute(self, event_data):
        self.validators.call(*event_data.args, **event_data.extended_kwargs)
        if not self._eval_cond(event_data):
            return False

        result = event_data.machine._activate(event_data)
        event_data.result = result
        return True



"""
Guess the number machine
========================

An StateMachine for the well know game.

Well leave the machine imagine a number and also play the game. Why not?

"""
import random

from statemachine import State
from statemachine import StateMachine


class GuessTheNumberMachine(StateMachine):

    start = State("Start", initial=True)
    low = State("Low")
    high = State("High")
    won = State("Won", final=True)
    lose = State("Lose", final=True)

    guess = (
        lose.from_(low, high, cond="max_guesses_reached")
        | won.from_(low, high, cond="guess_is_equal")
        | low.from_(low, high, start, cond="guess_is_lower")
        | high.from_(low, high, start, cond="guess_is_higher")
    )

    def __init__(self, max_attempts=5, lower=1, higher=5, seed=42):
        self.max_attempts = max_attempts
        self.lower = lower
        self.higher = higher
        self.guesses = 0

        # lets play a not so random game, or our tests will be crazy
        random.seed(seed)
        self.number = random.randint(self.lower, self.higher)
        super(GuessTheNumberMachine, self).__init__()

    def max_guesses_reached(self):
        return self.guesses >= self.max_attempts

    def before_guess(self, number):
        self.guesses += 1
        print("You guess is {}...".format(number))

    def guess_is_lower(self, number):
        return number < self.number

    def guess_is_higher(self, number):
        return number > self.number

    def guess_is_equal(self, number):
        return self.number == number

    def on_enter_start(self):
        print("(psss.. don't tell anyone the number is {})".format(self.number))
        print(
            "I'm thinking of a number between {} and {}. Can you guess what it is?".format(
                self.lower, self.higher
            )
        )

    def on_enter_low(self):
        print("Too low. Try again.")

    def on_enter_high(self):
        print("Too high. Try again.")

    def on_enter_won(self):
        print(
            "Congratulations, you guessed the number in {} guesses!".format(
                self.guesses
            )
        )

    def on_enter_lose(self):
        print("Oh, no! You've spent all your {} attempts!".format(self.guesses))


# %%
# Playing
# -------
#

sm = GuessTheNumberMachine(seed=103)

# %%

sm.guess(random.randint(1, 5))

# %%

sm

# %%

sm.guess(random.randint(1, 5))

# %%

sm.guess(random.randint(1, 5))


sm

# %%

# %%

sm.guess(random.randint(1, 5))

# %%

sm.guess(random.randint(1, 5))

# %%

sm

# %%

try:
    sm.guess(random.randint(1, 5))
except Exception as e:
    print(e)



# coding: utf-8
import warnings

from .utils import qualname


_REGISTRY = {}
_initialized = False


def register(cls):
    _REGISTRY[qualname(cls)] = cls
    _REGISTRY[cls.__name__] = cls
    return cls


def get_machine_cls(name):
    init_registry()
    if "." not in name:
        warnings.warn(
            """Use fully qualified names (<module>.<class>) for state machine mixins.""",
            DeprecationWarning,
        )
    return _REGISTRY[name]


def init_registry():
    global _initialized
    if not _initialized:
        load_modules(["statemachine", "statemachines"])
        _initialized = True


def _has_django():
    try:
        import django  # noqa

        return True
    except ImportError:
        # Not a django project
        pass
    return False


def _autodiscover_modules(module_name):  # pragma: no cover
    "Django 1.6 compat to provide `autodiscover_modules`"
    from django.conf import settings
    from django.utils.importlib import import_module

    for app in settings.INSTALLED_APPS:
        # Attempt to import the app's `module_name`.
        try:
            import_module("{app}.{module}".format(app=app, module=module_name))
        except Exception:
            pass


def load_modules(modules=None):
    if not _has_django():
        return
    try:  # pragma: no cover
        from django.utils.module_loading import autodiscover_modules
    except ImportError:  # pragma: no cover
        autodiscover_modules = _autodiscover_modules

    for module in modules:
        autodiscover_modules(module)



# coding: utf-8
from .utils import ensure_iterable


class Events(object):
    """A collection of event names."""

    def __init__(self):
        self.items = []

    def __repr__(self):
        sep = " " if len(self.items) > 1 else ""
        return sep.join(item for item in self.items)

    def __iter__(self):
        return iter(self.items)

    def add(self, events):
        if events is None:
            return self

        unprepared = ensure_iterable(events)
        for events in unprepared:
            for event in events.split(" "):
                if event in self.items:
                    continue
                self.items.append(event)

        return self

    def match(self, event):
        return any(t == event for t in self)



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import mock
import pytest


@pytest.fixture()
def event_mock():
    return mock.MagicMock()


@pytest.fixture()
def traffic_light_machine(event_mock):
    from statemachine import StateMachine, State

    class TrafficLightMachineStateEvents(StateMachine):
        "A traffic light machine"
        green = State("Green", initial=True)
        yellow = State("Yellow")
        red = State("Red")

        cycle = green.to(yellow) | yellow.to(red) | red.to(green)

        def on_enter_state(self, event_data):
            event_mock.on_enter_state(event_data.transition.target)

        def on_exit_state(self, event_data):
            event_mock.on_exit_state(event_data.state)

        def on_enter_green(self):
            event_mock.on_enter_green(self)

        def on_exit_green(self):
            event_mock.on_exit_green(self)

        def on_enter_yellow(self):
            event_mock.on_enter_yellow(self)

        def on_exit_yellow(self):
            event_mock.on_exit_yellow(self)

        def on_enter_red(self):
            event_mock.on_enter_red(self)

        def on_exit_red(self):
            event_mock.on_exit_red(self)

    return TrafficLightMachineStateEvents


class TestStateCallbacks(object):
    def test_should_call_on_enter_generic_state(
        self, event_mock, traffic_light_machine
    ):
        machine = traffic_light_machine()
        machine.cycle()
        event_mock.on_enter_state.call_args_list == [
            mock.call(machine.green),
            mock.call(machine.yellow),
        ]

    def test_should_call_on_exit_generic_state(self, event_mock, traffic_light_machine):
        machine = traffic_light_machine()
        machine.cycle()
        event_mock.on_exit_state.assert_called_once_with(machine.green)

    def test_should_call_on_enter_of_specific_state(
        self, event_mock, traffic_light_machine
    ):
        machine = traffic_light_machine()
        machine.cycle()
        event_mock.on_enter_yellow.assert_called_once_with(machine)

    def test_should_call_on_exit_of_specific_state(
        self, event_mock, traffic_light_machine
    ):
        machine = traffic_light_machine()
        machine.cycle()
        event_mock.on_exit_green.assert_called_once_with(machine)

    def test_should_be_on_the_previous_state_when_exiting(
        self, event_mock, traffic_light_machine
    ):
        machine = traffic_light_machine()

        def assert_is_green_from_state(s):
            assert s.value == "green"

        def assert_is_green(m):
            assert m.green.is_active

        event_mock.on_exit_state.side_effect = assert_is_green_from_state
        event_mock.on_exit_green.side_effect = assert_is_green

        machine.cycle()

    def test_should_be_on_the_next_state_when_entering(
        self, event_mock, traffic_light_machine
    ):
        machine = traffic_light_machine()

        def assert_is_yellow_from_state(s):
            assert s.value == "yellow"

        def assert_is_yellow(m):
            assert m.yellow.is_active

        event_mock.on_enter_state.side_effect = assert_is_yellow_from_state
        event_mock.on_enter_yellow.side_effect = assert_is_yellow

        machine.cycle()



# coding: utf-8
import inspect
from collections import namedtuple
from functools import wraps
from operator import attrgetter

from .exceptions import AttrNotFound
from .utils import ugettext as _


class ObjectConfig(namedtuple("ObjectConfig", "obj skip_attrs")):
    """Configuration for objects passed to resolver_factory.

    Args:
        obj: Any object that will serve as lookup for attributes.
        skip_attrs: Protected attrs that will be ignored on the search.
    """

    @classmethod
    def from_obj(cls, obj):
        if isinstance(obj, ObjectConfig):
            return obj
        else:
            return cls(obj, set())


def methodcaller(method):
    """Build a wrapper that adapts the received arguments to the inner method signature"""

    # spec is a named tuple ArgSpec(args, varargs, keywords, defaults)
    # args is a list of the argument names (it may contain nested lists)
    # varargs and keywords are the names of the * and ** arguments or None
    # defaults is a tuple of default argument values or None if there are no default arguments
    spec = inspect.getargspec(method)
    keywords = spec.keywords
    expected_args = list(spec.args)
    expected_kwargs = spec.defaults or {}

    # discart "self" argument for bounded methods
    if hasattr(method, "__self__") and expected_args and expected_args[0] == "self":
        expected_args = expected_args[1:]

    @wraps(method)
    def wrapper(*args, **kwargs):
        if spec.varargs is not None:
            filtered_args = args
        else:
            filtered_args = [
                kwargs.get(k, (args[idx] if idx < len(args) else None))
                for idx, k in enumerate(expected_args)
            ]

        if keywords is not None:
            filtered_kwargs = kwargs
        else:
            filtered_kwargs = {k: v for k, v in kwargs.items() if k in expected_kwargs}

        return method(*filtered_args, **filtered_kwargs)

    return wrapper


def _get_func_by_attr(attr, *configs):
    for config in configs:
        if attr in config.skip_attrs:
            continue
        func = getattr(config.obj, attr, None)
        if func is not None:
            break
    else:
        raise AttrNotFound(
            _("Did not found name '{}' from model or statemachine").format(attr)
        )
    return func, config.obj


def ensure_callable(attr, *objects):
    """Ensure that `attr` is a callable, if not, tries to retrieve one from any of the given
    `objects`.

    Args:
        attr (str or callable): A property/method name or a callable.
        objects: A list of objects instances that will serve as lookup for the given attr.
            The result `callable`, if any, will be a wrapper to the first object's attr that
            has the given ``attr``.
    """
    if callable(attr) or isinstance(attr, property):
        return methodcaller(attr)

    # Setup configuration if not present to normalize the internal API
    configs = [ObjectConfig.from_obj(obj) for obj in objects]

    func, obj = _get_func_by_attr(attr, *configs)

    if not callable(func):
        # if `attr` is not callable, then it's an attribute or property,
        # so `func` contains it's current value.
        # we'll build a method that get's the fresh value for each call
        getter = attrgetter(attr)

        def wrapper(*args, **kwargs):
            return getter(obj)

        return wrapper

    return methodcaller(func)


def resolver_factory(*objects):
    """Factory that returns a configured resolver."""

    @wraps(ensure_callable)
    def wrapper(attr):
        return ensure_callable(attr, *objects)

    return wrapper



class TestDeprecatedAttrs:
    def test_event_identifier(self, campaign_machine):
        sm = campaign_machine()
        assert [e.identifier for e in sm.events] == [e.name for e in sm.events]

    def test_state_identifier(self, campaign_machine):
        sm = campaign_machine()
        assert sm.draft.identifier == sm.draft.id

    def test_transitions_list(self, campaign_machine):
        sm = campaign_machine()
        assert [t.identifier for t in sm.transitions] == [e.name for e in sm.events]

    def test_transition_identifier(self, campaign_machine):
        sm = campaign_machine()
        assert [t.identifier for t in sm.draft.transitions] == ["add_job", "produce"]

    def test_allowed_transitions(self, campaign_machine):
        sm = campaign_machine()
        assert sm.allowed_transitions == sm.allowed_events

    def test_class_attr_transitions(self, campaign_machine):
        assert sorted(t.identifier for t in campaign_machine.transitions) == sorted(
            ["add_job", "produce", "deliver"]
        )

    def test_run(self, campaign_machine):
        sm = campaign_machine()
        sm.run("produce")
        assert sm.producing.is_active

    def test_is_stateid_check_property(self, campaign_machine):
        sm = campaign_machine()
        assert sm.is_draft == sm.draft.is_active
        sm.produce()
        assert sm.is_producing == sm.producing.is_active



# coding: utf-8


class TestActions:
    def test_should_return_all_before_results(self, AllActionsMachine):
        import tests.examples.all_actions_machine  # noqa

    def test_should_allow_actions_on_the_model(self):
        # just importing, as the example has assertions
        import tests.examples.order_control_rich_model_machine  # noqa

    def test_should_ignore_preconfigured_callbacks_not_present_(self, campaign_machine):
        sm = campaign_machine()
        assert sm.draft.enter.items == []
        assert sm.draft.exit.items == []



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

from .statemachine import State
from .statemachine import StateMachine

__author__ = """Fernando Macedo"""
__email__ = "fgmacedo@gmail.com"
__version__ = "1.0.1"

__all__ = ["StateMachine", "State"]



import warnings

from .event_data import EventData
from .exceptions import TransitionNotAllowed


class Event(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.name)

    def __call__(self, machine, *args, **kwargs):
        return self.trigger(machine, *args, **kwargs)

    def trigger(self, machine, *args, **kwargs):
        event_data = EventData(machine, self.name, *args, **kwargs)

        def trigger_wrapper():
            """Wrapper that captures event_data as closure."""
            return self._trigger(event_data)

        return machine._process(trigger_wrapper)

    def _trigger(self, event_data):
        event_data.source = event_data.machine.current_state
        event_data.state = event_data.machine.current_state
        event_data.model = event_data.machine.model

        try:
            self._process(event_data)
        except Exception as error:
            event_data.error = error
            # TODO: Log errors
            # TODO: Allow exception handlers
            raise
        return event_data.result

    def _process(self, event_data):
        for transition in event_data.source.transitions:
            if not transition.match(event_data.event):
                continue
            event_data._set_transition(transition)
            if transition.execute(event_data):
                event_data.executed = True
                break
        else:
            raise TransitionNotAllowed(event_data.event, event_data.state)

    @property
    def identifier(self):
        warnings.warn(
            "identifier is deprecated. Use `name` instead", DeprecationWarning
        )
        return self.name


def trigger_event_factory(event):
    """Build a method that sends specific `event` to the machine"""
    event_instance = Event(event)

    def trigger_event(self, *args, **kwargs):
        return event_instance(self, *args, **kwargs)

    trigger_event.name = event
    trigger_event.identifier = event

    return trigger_event



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest

from .models import MyModel
from statemachine import State
from statemachine import StateMachine
from statemachine.statemachine import Transition


def test_transition_representation(campaign_machine):
    s = repr([t for t in campaign_machine.draft.transitions if t.event == "produce"][0])
    assert s == (
        "Transition("
        "State('Draft', id='draft', value='draft', initial=True, final=False), "
        "State('Being produced', id='producing', value='producing', "
        "initial=False, final=False), event='produce')"
    )


def test_list_machine_events(classic_traffic_light_machine):
    machine = classic_traffic_light_machine()
    transitions = [t.name for t in machine.events]
    assert transitions == ["go", "slowdown", "stop"]


def test_list_state_transitions(classic_traffic_light_machine):
    machine = classic_traffic_light_machine()
    events = [t.event for t in machine.green.transitions]
    assert events == ["slowdown"]


def test_transition_should_accept_decorator_syntax(traffic_light_machine):
    machine = traffic_light_machine()
    assert machine.current_state == machine.green


def test_transition_as_decorator_should_call_method_before_activating_state(
    traffic_light_machine,
):
    machine = traffic_light_machine()
    assert machine.current_state == machine.green
    assert (
        machine.cycle(1, 2, number=3, text="x") == "Running cycle from green to yellow"
    )
    assert machine.current_state == machine.yellow


@pytest.mark.parametrize(
    "machine_name",
    [
        "traffic_light_machine",
        "reverse_traffic_light_machine",
    ],
)
def test_cycle_transitions(request, machine_name):
    machine_class = request.getfixturevalue(machine_name)
    machine = machine_class()
    expected_states = ["green", "yellow", "red"] * 2
    for expected_state in expected_states:
        assert machine.current_state.id == expected_state
        machine.cycle()


def test_transition_call_can_only_be_used_as_decorator():
    source, dest = State("Source"), State("Destination")
    transition = Transition(source, dest)

    with pytest.raises(TypeError):
        transition("not a callable")


@pytest.fixture(params=["bounded", "unbounded"])
def transition_callback_machine(request):
    if request.param == "bounded":

        class ApprovalMachine(StateMachine):
            "A workflow"
            requested = State("Requested", initial=True)
            accepted = State("Accepted")

            validate = requested.to(accepted)

            def on_validate(self):
                self.model.calls.append("on_validate")
                return "accepted"

    elif request.param == "unbounded":

        class ApprovalMachine(StateMachine):
            "A workflow"
            requested = State("Requested", initial=True)
            accepted = State("Accepted")

            @requested.to(accepted)
            def validate(self):
                self.model.calls.append("on_validate")
                return "accepted"

    else:
        raise ValueError("machine not defined")

    return ApprovalMachine


def test_statemachine_transition_callback(transition_callback_machine):
    model = MyModel(state="requested", calls=[])
    machine = transition_callback_machine(model)
    assert machine.validate() == "accepted"
    assert model.calls == ["on_validate"]


def test_can_run_combined_transitions():
    class CampaignMachine(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True)
        producing = State("Being produced")
        closed = State("Closed")

        abort = draft.to(closed) | producing.to(closed) | closed.to(closed)
        produce = draft.to(producing)

    machine = CampaignMachine()

    machine.abort()

    assert machine.closed.is_active


def test_transitions_to_the_same_estate_as_itself():
    class CampaignMachine(StateMachine):
        "A workflow machine"
        draft = State("Draft", initial=True)
        producing = State("Being produced")
        closed = State("Closed")

        update = draft.to.itself()
        abort = draft.to(closed) | producing.to(closed) | closed.to.itself()
        produce = draft.to(producing)

    machine = CampaignMachine()

    machine.update()

    assert machine.draft.is_active


class TestReverseTransition(object):
    @pytest.mark.parametrize(
        "initial_state",
        [
            "green",
            "yellow",
            "red",
        ],
    )
    def test_reverse_transition(self, reverse_traffic_light_machine, initial_state):
        machine = reverse_traffic_light_machine(start_value=initial_state)
        assert machine.current_state.id == initial_state

        machine.stop()

        assert machine.red.is_active


def test_should_transition_with_a_dict_as_return():
    "regression test that verifies if a dict can be used as return"

    expected_result = {
        "a": 1,
        "b": 2,
        "c": 3,
    }

    class ApprovalMachine(StateMachine):
        "A workflow"
        requested = State("Requested", initial=True)
        accepted = State("Accepted")
        rejected = State("Rejected")

        accept = requested.to(accepted)
        reject = requested.to(rejected)

        def on_accept(self):
            return expected_result

    machine = ApprovalMachine()

    result = machine.send("accept")
    assert result == expected_result



# pragma: no cover
from .factory import StateMachineMetaclass
from .statemachine import BaseStateMachine


class StateMachine(BaseStateMachine):  # pragma: no cover
    __metaclass__ = StateMachineMetaclass



# coding: utf-8
from __future__ import absolute_import
from __future__ import unicode_literals

from .utils import ugettext as _


class StateMachineError(Exception):
    "Base exception for this project, all exceptions that can be raised inherit from this class."


class InvalidDefinition(StateMachineError):
    "The state machine has a definition error"


class InvalidStateValue(InvalidDefinition):
    "The current model state value is not mapped to a state definition."

    def __init__(self, value):
        self.value = value
        msg = _("{!r} is not a valid state value.").format(value)
        super(InvalidStateValue, self).__init__(msg)


class AttrNotFound(InvalidDefinition):
    "There's no method or property with the given name"


class TransitionNotAllowed(StateMachineError):
    "The transition can't run from the current state."

    def __init__(self, event, state):
        self.event = event
        self.state = state
        msg = _("Can't {} when in {}.").format(self.event, self.state.name)
        super(TransitionNotAllowed, self).__init__(msg)



from collections import OrderedDict

from .utils import ensure_iterable


class TransitionList(object):
    def __init__(self, transitions=None):
        self.transitions = list(transitions) if transitions else []

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.transitions)

    def __or__(self, other):
        return TransitionList(self.transitions).add_transitions(other)

    def add_transitions(self, transition):
        if isinstance(transition, TransitionList):
            transition = transition.transitions
        transitions = ensure_iterable(transition)

        for transition in transitions:
            self.transitions.append(transition)

        return self

    def __getitem__(self, index):
        return self.transitions[index]

    def __len__(self):
        return len(self.transitions)

    def _add_callback(self, callback, name, is_event=False):
        for transition in self.transitions:
            list_obj = getattr(transition, name)
            list_obj._add_unbounded_callback(
                callback,
                is_event=is_event,
                transitions=self,
            )
        return callback

    def __call__(self, f):
        return self._add_callback(f, "on", is_event=True)

    def before(self, f):
        return self._add_callback(f, "before")

    def after(self, f):
        return self._add_callback(f, "after")

    def on(self, f):
        return self._add_callback(f, "on")

    def add_event(self, event):
        for transition in self.transitions:
            transition.add_event(event)

    @property
    def unique_events(self):
        # Compat Python2.7: Using OrderedDict to get a unique ordered list
        tmp_list = OrderedDict()
        for transition in self.transitions:
            for event in transition.events:
                tmp_list[event] = True

        return list(tmp_list.keys())



"""
Order control machine (rich model)
==================================

An StateMachine that demonstrates :ref:`Actions` being used on a rich model.

"""
from statemachine import State
from statemachine import StateMachine
from statemachine.exceptions import AttrNotFound


class Order(object):
    def __init__(self):
        self.order_total = 0
        self.payments = []
        self.payment_received = False

    def payments_enough(self, amount):
        return sum(self.payments) + amount >= self.order_total

    def add_to_order(self, amount):
        self.order_total += amount
        return self.order_total

    def on_receive_payment(self, amount):
        self.payments.append(amount)
        return self.payments

    def after_receive_payment(self):
        self.payment_received = True

    def wait_for_payment(self):
        self.payment_received = False


class OrderControl(StateMachine):
    waiting_for_payment = State(
        "Waiting for payment", initial=True, enter="wait_for_payment"
    )
    processing = State("Processing")
    shipping = State("Shipping")
    completed = State("Completed", final=True)

    add_to_order = waiting_for_payment.to(waiting_for_payment, before="add_to_order")
    receive_payment = waiting_for_payment.to(
        processing, cond="payments_enough"
    ) | waiting_for_payment.to(waiting_for_payment, unless="payments_enough")
    process_order = processing.to(shipping, cond="payment_received")
    ship_order = shipping.to(completed)


# %%
# Testing
# -------
#
# Let's first try to create a statemachine instance, using the default dummy model that don't have
# the needed methods to complete the state machine. Since the required methods will not be found
# either in the state machine or in the model, an exception ``AttrNotFound`` will be raised.

try:
    control = OrderControl()
except AttrNotFound as e:
    assert str(e) == "Did not found name 'payment_received' from model or statemachine"

# %%
# Now initializing with a proper ``order`` instance.

order = Order()
control = OrderControl(order)

# %%
# Send events to add to order

assert control.send("add_to_order", 3) == 3
assert control.send("add_to_order", 7) == 10

# %%
# Receive a payment of $4...

control.send("receive_payment", 4)

# %%
# Since there's still $6 left to fulfill the payment, we cannot process the order.
try:
    control.send("process_order")
except StateMachine.TransitionNotAllowed as err:
    print(err)

# %%

control

# %%
# Now paying the left amount, we can proceed.

control.send("receive_payment", 6)

# %%

control

# %%

control.send("process_order")

# %%

control

# %%

control.send("ship_order")

# %%
# Just checking the final expected state

order.order_total

# %%

order.payments

# %%

control.completed.is_active

# %%

control


# %%
assert order.order_total == 10
assert order.payments == [4, 6]
assert control.completed.is_active



# coding: utf-8
import pytest

from statemachine import exceptions


@pytest.fixture
def BaseMachine():
    from statemachine import StateMachine, State

    class BaseMachine(StateMachine):
        state_1 = State("1", initial=True)
        state_2 = State("2")
        trans_1_2 = state_1.to(state_2)

    return BaseMachine


@pytest.fixture
def InheritedClass(BaseMachine):
    class InheritedClass(BaseMachine):
        pass

    return InheritedClass


@pytest.fixture
def ExtendedClass(BaseMachine):
    from statemachine import State

    class ExtendedClass(BaseMachine):
        state_3 = State("3")
        trans_2_3 = BaseMachine.state_2.to(state_3)

    return ExtendedClass


@pytest.fixture
def OverridedClass(BaseMachine):
    from statemachine import State

    class OverridedClass(BaseMachine):
        state_2 = State("2", value="state_2")

        trans_1_2 = BaseMachine.state_1.to(state_2)

    return OverridedClass


@pytest.fixture
def OverridedTransitionClass(BaseMachine):
    from statemachine import State

    class OverridedTransitionClass(BaseMachine):
        state_3 = State("3")
        trans_1_2 = BaseMachine.state_1.to(state_3)

    return OverridedTransitionClass


def test_should_inherit_states_and_transitions(BaseMachine, InheritedClass):
    assert InheritedClass.states == [
        BaseMachine.state_1,
        BaseMachine.state_2,
    ]

    expected = [e.name for e in BaseMachine.events]
    actual = [e.name for e in InheritedClass.events]
    assert actual == expected


def test_should_extend_states_and_transitions(BaseMachine, ExtendedClass):
    assert ExtendedClass.states == [
        BaseMachine.state_1,
        BaseMachine.state_2,
        ExtendedClass.state_3,
    ]

    base_events = [e.name for e in BaseMachine.events]
    expected = base_events + [ExtendedClass.trans_2_3.name]
    actual = [e.name for e in ExtendedClass.events]
    assert actual == expected


def test_should_execute_transitions(ExtendedClass):
    instance = ExtendedClass()
    instance.trans_1_2()
    instance.trans_2_3()

    assert instance.state_3.is_active


@pytest.mark.xfail(reason="State overriding is not supported")
def test_dont_support_overriden_states(OverridedClass):
    # There's no support for overriding states
    with pytest.raises(exceptions.InvalidDefinition):
        OverridedClass()


@pytest.mark.xfail(reason="Transition overriding is not supported")
def test_support_override_transitions(OverridedTransitionClass):
    instance = OverridedTransitionClass()

    instance.trans_1_2()
    assert instance.state_3.is_active



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import io

from setuptools import setup

with io.open("README.rst", encoding="utf-8") as readme_file:
    readme = readme_file.read()

long_description = readme

extras_require = {"diagrams": ["pydot"]}

setup(
    name="python-statemachine",
    version="1.0.1",
    description="Python Finite State Machines made easy.",
    long_description=long_description,
    author="Fernando Macedo",
    author_email="fgmacedo@gmail.com",
    url="https://github.com/fgmacedo/python-statemachine",
    packages=[
        "statemachine",
        "statemachine.contrib",
    ],
    package_dir={"statemachine": "statemachine"},
    extras_require=extras_require,
    include_package_data=True,
    license="MIT license",
    zip_safe=False,
    keywords="statemachine",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
    ],
    test_suite="tests",
)



# coding: utf-8
import mock
import pytest

from statemachine import State
from statemachine import StateMachine
from statemachine.callbacks import Callbacks
from statemachine.callbacks import CallbackWrapper
from statemachine.dispatcher import resolver_factory
from statemachine.exceptions import InvalidDefinition


@pytest.fixture
def ObjectWithCallbacks():
    class ObjectWithCallbacks(object):
        def __init__(self):
            super(ObjectWithCallbacks, self).__init__()
            self.name = "statemachine"
            self.callbacks = Callbacks(resolver=resolver_factory(self)).add(
                ["life_meaning", "name", "a_method"]
            )

        @property
        def life_meaning(self):
            return 42

        def a_method(self, *args, **kwargs):
            return args, kwargs

    return ObjectWithCallbacks


class TestCallbacksMachinery:
    def test_raises_exception_without_setup_phase(self):
        func = mock.Mock()

        callbacks = Callbacks()
        callbacks.add(func)

        with pytest.raises(InvalidDefinition):
            callbacks.call(1, 2, 3, a="x", b="y")

        func.assert_not_called()

    def test_can_add_callback(self):
        callbacks = Callbacks()
        func = mock.Mock()

        callbacks.add(func)
        callbacks.setup(lambda x: x)

        callbacks.call(1, 2, 3, a="x", b="y")

        func.assert_called_once_with(1, 2, 3, a="x", b="y")

    def test_callback_wrapper_is_hashable(self):
        wrapper = CallbackWrapper("something")
        set().add(wrapper)

    def test_can_add_callback_that_is_a_string(self):
        callbacks = Callbacks()
        func = mock.Mock()
        resolver = mock.Mock(return_value=func)

        callbacks.add("my_method").add("other_method")
        callbacks.add("last_one")
        callbacks.setup(resolver)

        callbacks.call(1, 2, 3, a="x", b="y")

        resolver.assert_has_calls(
            [
                mock.call("my_method"),
                mock.call("other_method"),
            ]
        )
        assert func.call_args_list == [
            mock.call(1, 2, 3, a="x", b="y"),
            mock.call(1, 2, 3, a="x", b="y"),
            mock.call(1, 2, 3, a="x", b="y"),
        ]

    def test_callbacks_are_iterable(self):
        callbacks = Callbacks()

        callbacks.add("my_method").add("other_method")
        callbacks.add("last_one")

        assert [c.func for c in callbacks] == ["my_method", "other_method", "last_one"]

    def test_add_many_callbacks_at_once(self):
        callbacks = Callbacks()
        method_names = ["my_method", "other_method", "last_one"]

        callbacks.add(method_names)

        assert [c.func for c in callbacks] == method_names

    @pytest.mark.parametrize("suppress_errors", [False, True])
    def test_raise_error_if_didnt_found_attr(self, suppress_errors):
        callbacks = Callbacks(resolver_factory(object()))

        if suppress_errors:
            callbacks.add("this_does_no_exist", suppress_errors=suppress_errors)
        else:
            with pytest.raises(InvalidDefinition):
                callbacks.add("this_does_no_exist", suppress_errors=suppress_errors)

    def test_collect_results(self):
        callbacks = Callbacks()
        func1 = mock.Mock(return_value=10)
        func2 = mock.Mock(return_value=("a", True))
        func3 = mock.Mock(return_value={"key": "value"})

        callbacks.add([func1, func2, func3])
        callbacks.setup(lambda x: x)

        results = callbacks.call(1, 2, 3, a="x", b="y")

        assert results == [
            10,
            ("a", True),
            {"key": "value"},
        ]

    def test_callbacks_values_resolution(self, ObjectWithCallbacks):
        x = ObjectWithCallbacks()
        assert x.callbacks.call(xablau=True) == [
            42,
            "statemachine",
            ((), {"xablau": True}),
        ]


class TestCallbacksAsDecorator:
    def test_decorate_unbounded_function(self, ObjectWithCallbacks):
        x = ObjectWithCallbacks()

        @x.callbacks
        def hero_lowercase(hero):
            return hero.lower()

        @x.callbacks
        def race_uppercase(race):
            return race.upper()

        assert x.callbacks.call(hero="Gandalf", race="Maia") == [
            42,
            "statemachine",
            ((), {"hero": "Gandalf", "race": "Maia"}),
            "gandalf",
            "MAIA",
        ]

        assert race_uppercase("Hobbit") == "HOBBIT"

    def test_decorate_unbounded_machine_methods(self):
        class MiniHeroJourneyMachine(StateMachine):

            ordinary_world = State("Ordinary World", initial=True)
            call_to_adventure = State("Call to Adventure")
            refusal_of_call = State("Refusal of the Call")

            adventure_called = ordinary_world.to(call_to_adventure)

            def __init__(self, *args, **kwargs):
                self.spy = mock.Mock(side_effect=lambda *x: x)
                super(MiniHeroJourneyMachine, self).__init__(*args, **kwargs)

            @ordinary_world.enter
            def enter_ordinary_world(self):
                """This is the hero's life before they begin their journey. It is their "normal"
                world, where they are comfortable and familiar.
                """
                self.spy("enter_ordinary_world")

            @call_to_adventure.enter
            def enter_call_to_adventure(self, request):
                """Something happens that forces the hero to leave their ordinary world and embark
                on a journey. This might be a direct call, like a prophecy or a request for help,
                or it might be a more subtle nudge, like a feeling of restlessness or a sense of
                something missing in their life."""
                self.spy("call_to_adventure", request)

            @ordinary_world.to(refusal_of_call)
            def refuse_call(self, reason):
                self.spy("refuse_call", reason)

        sm = MiniHeroJourneyMachine()
        sm.adventure_called(request="The darkness is comming")
        assert sm.spy.call_args_list == [
            mock.call("enter_ordinary_world"),
            mock.call("call_to_adventure", "The darkness is comming"),
        ]

        sm = MiniHeroJourneyMachine()
        sm.refuse_call(reason="Not prepared yet")
        assert sm.spy.call_args_list == [
            mock.call("enter_ordinary_world"),
            mock.call("refuse_call", "Not prepared yet"),
        ]



# coding: utf-8
import sys
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .dispatcher import ObjectConfig
from .dispatcher import resolver_factory
from .event import Event
from .event_data import EventData
from .exceptions import InvalidStateValue
from .exceptions import TransitionNotAllowed
from .model import Model
from .state import State
from .transition import Transition


class BaseStateMachine(object):

    TransitionNotAllowed = TransitionNotAllowed  # shortcut for handling exceptions

    _events = {}  # type: Dict[Any, Any]
    states = []  # type: List[State]
    states_map = {}  # type: Dict[Any, State]

    def __init__(self, model=None, state_field="state", start_value=None):
        self.model = model if model else Model()
        self.state_field = state_field
        self.start_value = start_value

        initial_transition = Transition(None, None, event="__initial__")
        self._setup(initial_transition)
        self._activate_initial_state(initial_transition)

    def __repr__(self):
        return "{}(model={!r}, state_field={!r}, current_state={!r})".format(
            type(self).__name__,
            self.model,
            self.state_field,
            self.current_state.id if self.current_state else None,
        )

    def _activate_initial_state(self, initial_transition):

        current_state_value = (
            self.start_value if self.start_value else self.initial_state.value
        )
        if self.current_state_value is None:

            try:
                initial_state = self.states_map[current_state_value]
            except KeyError:
                raise InvalidStateValue(current_state_value)

            # send an one-time event `__initial__` to enter the current state.
            # current_state = self.current_state
            initial_transition.target = initial_state
            initial_transition.before.clear()
            initial_transition.on.clear()
            initial_transition.after.clear()
            event_data = EventData(
                self,
                initial_transition.event,
                transition=initial_transition,
            )
            self._activate(event_data)

    def _get_protected_attrs(self):
        return (
            {
                "_abstract",
                "model",
                "state_field",
                "start_value",
                "initial_state",
                "final_states",
                "states",
                "_events",
                "states_map",
                "send",
            }
            | {s.id for s in self.states}
            | {e for e in self._events.keys()}
        )

    def _visit_states_and_transitions(self, visitor):
        for state in self.states:
            visitor(state)
            for transition in state.transitions:
                visitor(transition)

    def _setup(self, initial_transition):
        machine = ObjectConfig(self, skip_attrs=self._get_protected_attrs())
        model = ObjectConfig(self.model, skip_attrs={self.state_field})
        default_resolver = resolver_factory(machine, model)

        initial_transition._setup(default_resolver)
        self._visit_states_and_transitions(lambda x: x._setup(default_resolver))
        self.add_observer(machine, model)

    def add_observer(self, *observers):
        resolvers = [resolver_factory(ObjectConfig.from_obj(o)) for o in observers]
        self._visit_states_and_transitions(lambda x: x._add_observer(*resolvers))
        return self

    def _repr_html_(self):
        return '<div class="statemachine">{}</div>'.format(self._repr_svg_())

    def _repr_svg_(self):
        return self._graph().create_svg().decode()

    def _graph(self):
        from .contrib.diagram import DotGraphMachine

        return DotGraphMachine(self).get_graph()

    @property
    def current_state_value(self):
        value = getattr(self.model, self.state_field, None)
        return value

    @current_state_value.setter
    def current_state_value(self, value):
        if value not in self.states_map:
            raise InvalidStateValue(value)
        setattr(self.model, self.state_field, value)

    @property
    def current_state(self):
        # type: () -> Optional[State]
        return self.states_map.get(self.current_state_value, None)

    @current_state.setter
    def current_state(self, value):
        self.current_state_value = value.value

    @property
    def transitions(self):
        warnings.warn(
            "Property `transitions` is deprecated. Use `events` instead.",
            DeprecationWarning,
        )
        return self.events

    @property
    def events(self):
        return self.__class__.events

    @property
    def allowed_transitions(self):
        "get the callable proxy of the current allowed transitions"
        warnings.warn(
            "`allowed_transitions` is deprecated. Use `allowed_events` instead.",
            DeprecationWarning,
        )
        return [
            getattr(self, event)
            for event in self.current_state.transitions.unique_events
        ]

    @property
    def allowed_events(self):
        "get the callable proxy of the current allowed events"
        return [
            getattr(self, event)
            for event in self.current_state.transitions.unique_events
        ]

    def _process(self, trigger):
        """This method will also handle execution queue"""
        return trigger()

    def _activate(self, event_data):
        transition = event_data.transition
        source = event_data.state
        target = transition.target

        result = transition.before.call(*event_data.args, **event_data.extended_kwargs)
        if source is not None:
            source.exit.call(*event_data.args, **event_data.extended_kwargs)

        result += transition.on.call(*event_data.args, **event_data.extended_kwargs)

        self.current_state = target
        event_data.state = target

        target.enter.call(*event_data.args, **event_data.extended_kwargs)
        transition.after.call(*event_data.args, **event_data.extended_kwargs)

        if len(result) == 0:
            result = None
        elif len(result) == 1:
            result = result[0]

        return result

    def run(self, event, *args, **kwargs):
        warnings.warn(
            "`run` is deprecated. Use `send` instead.",
            DeprecationWarning,
        )
        event = Event(event)
        return event(self, *args, **kwargs)

    def send(self, event, *args, **kwargs):
        event = Event(event)
        return event(self, *args, **kwargs)


# Python 2
if sys.version_info[0] == 2:  # pragma: no cover
    from .factory_2 import StateMachine  # noqa
else:
    from .factory_3 import StateMachine  # noqa



EXPECTED_LOG = """Frodo on: draft--(add_job)-->draft
Frodo enter: draft from add_job
Frodo on: draft--(produce)-->producing
Frodo enter: producing from produce
"""


class TestObserver:
    def test_add_log_observer(self, campaign_machine, capsys):
        class LogObserver(object):
            def __init__(self, name):
                self.name = name

            def on_transition(self, event, state, target):
                print(
                    "{} on: {}--({})-->{}".format(self.name, state.id, event, target.id)
                )

            def on_enter_state(self, target, event):
                print("{} enter: {} from {}".format(self.name, target.id, event))

        sm = campaign_machine()

        sm.add_observer(LogObserver("Frodo"))

        sm.add_job()
        sm.produce()

        captured = capsys.readouterr()
        assert captured.out == EXPECTED_LOG



class EventData(object):
    def __init__(self, machine, event, *args, **kwargs):
        self.machine = machine
        self.event = event
        self.source = kwargs.get("source", None)
        self.state = kwargs.get("state", None)
        self.model = kwargs.get("model", None)
        self.executed = False
        self._set_transition(kwargs.get("transition", None))

        # runtime and error
        self.args = args
        self.kwargs = kwargs
        self.error = None
        self.result = None

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.__dict__)

    def _set_transition(self, transition):
        self.transition = transition
        self.target = getattr(transition, "target", None)

    @property
    def extended_kwargs(self):
        kwargs = self.kwargs.copy()
        kwargs["event_data"] = self
        kwargs["event"] = self.event
        kwargs["source"] = self.source
        kwargs["state"] = self.state
        kwargs["model"] = self.model
        kwargs["transition"] = self.transition
        kwargs["target"] = self.target
        return kwargs



# coding: utf-8
import pytest

from statemachine.contrib.diagram import DotGraphMachine
from statemachine.contrib.diagram import main


@pytest.fixture(
    params=[
        (
            "_repr_svg_",
            '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n<!DOCTYPE svg',
        ),
        (
            "_repr_html_",
            '<div class="statemachine"><?xml version="1.0" encoding="UTF-8" standalone=',
        ),
    ]
)
def expected_reprs(request):
    return request.param


@pytest.mark.parametrize(
    "machine_name",
    [
        "AllActionsMachine",
        "OrderControl",
    ],
)
def test_machine_repr_custom_(request, machine_name, expected_reprs):
    machine_cls = request.getfixturevalue(machine_name)
    machine = machine_cls()

    magic_method, expected_repr = expected_reprs
    repr = getattr(machine, magic_method)()
    assert repr.startswith(expected_repr)


def test_machine_dot(OrderControl):
    machine = OrderControl()

    graph = DotGraphMachine(machine)
    dot = graph()

    dot_str = dot.to_string()  # or dot.to_string()
    assert dot_str.startswith("digraph list {")


class TestDiagramCmdLine:
    def test_generate_image(self, tmp_path):
        out = tmp_path / "sm.svg"

        main(["tests.examples.traffic_light_machine.TrafficLightMachine", str(out)])

        assert out.read_text().startswith(
            '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n<!DOCTYPE svg'
        )

    def test_generate_complain_about_bad_sm_path(self, capsys, tmp_path):
        out = tmp_path / "sm.svg"

        with pytest.raises(ValueError) as e:
            main(
                [
                    "tests.examples.traffic_light_machine.TrafficLightMachineXXX",
                    str(out),
                ]
            )

        assert e.match("TrafficLightMachineXXX is not a subclass of StateMachine")



class MyModel(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return "{}({!r})".format(type(self).__name__, self.__dict__)
